from .base_hrm_page import BaseHrmPage
from .hrm_0_default_data_page import Hrm0DefaultDataPage
from .hrm_1_cumulative_operation_time_page import Hrm1CumulativeOperationTimePage
from .hrm_2_manufacturer_information_page import Hrm2ManufacturerInformationPage
from .hrm_3_product_information_page import Hrm3ProductInformationPage
from .hrm_4_previous_heart_beat_event_time_page import Hrm4PreviousHeartBeatEventTimePage
from .hrm_5_swim_interval_summary_page import Hrm5SwimIntervalSummaryPage
from .hrm_6_capabilities_page import Hrm6CapabilitiesPage
from .hrm_7_battery_status_page import Hrm7BatteryStatusPage
from .hrm_9_device_information_page import Hrm9DeviceInformationPage
from .hrm_32_hr_feature_page import Hrm32HrFeaturePage

__all__ = [
    "BaseHrmPage",
    "Hrm0DefaultDataPage",
    "Hrm1CumulativeOperationTimePage",
    "Hrm2ManufacturerInformationPage",
    "Hrm3ProductInformationPage",
    "Hrm4PreviousHeartBeatEventTimePage",
    "Hrm5SwimIntervalSummaryPage",
    "Hrm6CapabilitiesPage",
    "Hrm7BatteryStatusPage",
    "Hrm9DeviceInformationPage",
    "Hrm32HrFeaturePage",
]
